#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: backtest.py
Author: wuyiyang(wuyiyang@hcyjs.com), wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/15 11:18:01
"""
import os
import fqdatac
import logging
import pandas as pd
import numpy as np
from typing import List
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor, Future
from . import FactorBackTestException
from ..models.factor_model import BackTestConfig
from ..context import FatbullsFactorContext

"""基础变量"""
FIRST_DAY = 0
LAST_ONE_DAY = -1
LAST_ONE_MONTH = -22
LAST_THREE_MONTH = -64
LAST_ONE_YEAR = -245
LAST_THREE_YEAR = -490
LAST_FIVE_YEAR = -1225
DAYS_PER_YEAR = 244


"""
'output_dirpath': output_path,   # 可选
        'backtest_name': 'test_backtest',
        'sd': '20170101',
        'ed': '20201231',
        'task_num': 'WaitForOne',  # 可选
        'univ_name': 'cnall',      # 可选
        'fdays': 1,                # 可选
        'rettype': 'vwap30',       # 可选
        'directory_list': [
            '/data/quant/factors/DTech/DTech_volume',
            '/data/quant/factors/DTech/DTech_amt',
            '/data/quant/factors/DTech/DTech_high',
            '/data/quant/factors/DTech/DTech_ret',
            '/data/quant/factors/DTech/DTech_logret',
            '/data/quant/factors/DTech/DTech_opn',
            '/data/quant/factors/DTech/DTech_vwap'
        ],
"""


class FactorBackTestBase(object):
    """ 回测引擎基类 """

    def __init__(self):
        self.today = datetime.now().strftime("%Y%m%d")
        self.backtest_day = datetime.now().strftime("%Y%m%d")


class FactorBackTestEngine(FactorBackTestBase):
    """因子回测主引擎"""

    def __init__(self, backtest_config: BackTestConfig):
        """回测参数初始化"""
        FactorBackTestBase.__init__(self)
        self.start_date = backtest_config.sd
        self.end_date = backtest_config.ed
        self.task_num = backtest_config.task_num
        self.backtest_name = backtest_config.backtest_name
        self.univ_name = backtest_config.univ_name
        self.forward_days = backtest_config.fdays
        self.return_type = backtest_config.rettype
        self.factor_dirs = backtest_config.factor_dirs

        """Context"""
        self._fbctx = FatbullsFactorContext.get_instance()

        """回测fqdatac相关数据"""
        self.alldays = None
        self.univ = None
        self.cnall = None

        """ 回测数据容器 """
        self.factors = {}
        self.factor_data = {}

        """对照股票的return数据"""
        self.forward_return1 = None
        self.forward_return = None

    def initialize(self):
        """
        第一步：初始化文件检查
        """
        self._check_date()
        self._fbctx.set_alldays(fqdatac.get_daysrange(self.start_date, self.end_date))
        self._fbctx.set_univ_userselect(Universe(self.univ_name, self.start_date, self.end_date))
        self._fbctx.set_univ_cnall(Universe('cnall', self.start_date, self.end_date))

    def load_factors(self):
        """
        第二步： 数据加载预处理
        将因子load进container
        查路径里的文件是否齐全 是否重复
        """
        logging.info("start factor loading")
        for dirpath in self.factor_dirs:
            factor_name = os.path.basename(dirpath)
            factor = Factor(factor_name, dirpath)
            self.factors[factor_name] = factor
            self.factor_data[factor_name] = factor.load_data(use_cache=False)
            logging.info("factor:{} successfully loaded".format(factor_name))
        self._only_univ_filter()
        logging.info("start fwdret loading")
        self._load_fwdret()

    def _only_univ_filter(self):
        """
        只保留当天在universe中的股票进行测试
        """
        univ = self._fbctx.get_univ_userselect()
        for k, v in self.factor_data.items():
            for dt in list(v.keys()):
                v[dt] = v[dt][v[dt]["K"].isin(list(univ.data[dt]["K"]))]

    def _load_fwdret(self):
        """
        获取对照的股票return数据
        """
        self.forward_return = fqdatac.get_bfwd(
            sd=self.start_date,
            ed=self.end_date,
            univ_name=self.univ_name,
            ret_type=self.return_type,
            dayshift=self.forward_days,
        )
        if self.forward_days == 1:
            self.forward_return1 = self.forward_return.copy()
        else:
            self.forward_return1 = fqdatac.get_bfwd(
                sd=self.start_date,
                ed=self.end_date,
                univ_name=self.univ_name,
                ret_type=self.return_type,
                dayshift=1,
            )

    def run(self, max_workers=10):
        """回测主函数"""
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures: List[Future] = []
            for factor_name in self.factors.keys():
                factor_data = pd.concat(self.factor_data[factor_name].values(), ignore_index=True)
                factor_data = pd.merge(factor_data, self.forward_return, on=['K', 'D', 'T'], how='left')
                future = executor.submit(FactorCalculateTask(factor_name, factor_data))
                futures.append(future)
                del factor_data
            executor.shutdown(wait=True)
            self.save_report(futures)

    def save_report(self, futures: List[Future]):
        cols = ['Coverage', 'ICs', 'SCs']
        ls_cols = ['Summary', 'Turnover', 'LSreturns', 'Wealth', 'HitRate']
        sheet_names = cols + [f'top_{s}'for s in ls_cols] + [f'bottom_{s}' for s in ls_cols]
        combine_data = dict()
        for future in futures:
            backtest_result = future.result()
            for sheet_name in sheet_names:
                if 'Summary' in sheet_name:
                    if sheet_name not in combine_data:
                        combine_data[sheet_name] = list()
                    combine_data[sheet_name].append(backtest_result[sheet_name])
                else:
                    if sheet_name not in combine_data:
                        combine_data[sheet_name] = pd.DataFrame()
                    df_right = backtest_result[sheet_name].set_index(['D', 'T'])
                    combine_data[sheet_name] = pd.concat([combine_data[sheet_name], df_right], axis=1)
        report_path = self.get_report_path()
        file_name = "BT_factor_report_{}_{}_{}_{}_{}.xlsx".format(self.today, self.task_num, self.univ_name,
                                                                  self.start_date, self.end_date)
        file_name = "{}/{}".format(report_path, file_name)
        with pd.ExcelWriter(file_name) as writer:
            for sheet_name in combine_data:
                if 'Summary' in sheet_name:
                    df_sheet = pd.DataFrame.from_dict(combine_data[sheet_name])
                else:
                    df_sheet = combine_data[sheet_name].reset_index(['D', 'T'])
                df_sheet.to_excel(writer, sheet_name=sheet_name)

    def _check_date(self):
        try:
            assert fqdatac.get_day(self.end_date, -60) >= self.start_date
            assert self.end_date <= fqdatac.get_day(self.backtest_day, -1)
        except AssertionError:
            raise FactorBackTestException("invalid date range inputs")

    def get_report_path(self):
        report_path = self._fbctx.get_base_report_dirpath()
        report_path = os.path.join(report_path, self.task_num)
        if not os.path.exists(report_path):
            os.makedirs(report_path, exist_ok=True)
        return report_path


class FactorCalculateTask(object):
    """多线程执行引擎"""

    def __init__(self, factor_name: str, factor_data: pd.DataFrame):
        self.factor_name = factor_name
        self.factor_data = factor_data
        self._factor_config = FatbullsFactorContext.get_instance().get_factor_config()
        self.forward_return_name = list(factor_data.columns)[-1]
        self.factor_forward = self.factor_name + '_' + self.forward_return_name
        self.grouped = None
        self.df_longshort = None
        self.backtest_result = dict()
        self.coverage = dict()

    def _middle_calculate(self):
        self.factor_data = self.factor_data.sort_values(by=['D', 'T', 'K'], axis=0, ascending=True)
        self.grouped = self.factor_data.groupby(by=['D', 'T'], as_index=False)
        self.factor_data['q'] = self.factor_data.groupby(by=['D', 'T'], as_index=False)[self.factor_name].transform(
            lambda x: pd.qcut(x, 10, labels=["Q{}".format(i) for i in range(1, 11)]))
        self.factor_data['q'] = self.factor_data['q'].astype(str)
        self.df_longshort = self.factor_data.query("q == 'Q1' or q == 'Q10'")

    def __call__(self, *args, **kwargs):
        self.backtest_result['factor_name'] = self.factor_name
        self._middle_calculate() # noqa
        self._coverage_summary()
        self.calculate_ic()
        self.calculate_coverage()
        self.calculate_turnover()
        self.calculate_lsreturn()
        self.calculate_hitrate()
        self.calculate_wealth()
        # calculate_serial_corr和summary计算顺序不能反
        self.calculate_serial_corr()
        self.summary()
        self.summary(summary_type="top")
        self.summary(summary_type="bottom")
        del self.factor_data
        return self.backtest_result

    def calculate_ic(self):
        """4.1 计算IC相关性, 不区分LongShort"""
        df_ic = self.grouped.apply(lambda x: x[self.factor_name].corr(x[self.forward_return_name], 'spearman'))
        df_ic.columns = ['D', 'T', self.factor_forward]
        self.backtest_result['ICs'] = df_ic
        del df_ic

    def calculate_coverage(self):
        """4.1 计算coverage, 不区分LongShort"""
        df_coverage = self.grouped.agg({'K': 'count'})
        df_coverage = df_coverage.rename(columns={"K": self.factor_name})
        df_coverage = df_coverage[['D', 'T', self.factor_name]].rename(columns={self.factor_name: self.factor_forward})
        self.backtest_result['Coverage'] = df_coverage
        del df_coverage

    def calculate_turnover(self):
        """
            4.1 urnover , 分为LongShort、Long、Short
            LongShort Turnover = 1 - (C1 + C10)/(S1 + S10)
            Long Turnover  = 1 - C10/S10
            Short Turnover = 1 - C1/S1
        """
        self.backtest_result['Turnover'] = self._calculate_turnover(self.df_longshort)
        self.backtest_result['top_Turnover'] = self._calculate_turnover(self.df_longshort.query("q == 'Q1'"))
        self.backtest_result['bottom_Turnover'] = self._calculate_turnover(self.df_longshort.query("q == 'Q10'"))

    def _calculate_turnover(self, df_q: pd.DataFrame):

        def _calc_turnover(x, y):
            """计算交集"""
            if x is None or y is None:
                return 1.0
            return 1.0 - len(set(x).intersection(set(y))) / (len(x) + len(y))

        df_q = df_q.groupby(['D', 'T'], as_index=False).agg({"K": list})
        df_q.columns = ['D', 'T', 'K_list']
        df_q['K_list_prev'] = df_q.sort_values(by=['D', 'T'], ascending=True)['K_list'].shift(1)
        df_q[self.factor_name] = df_q.apply(lambda x: _calc_turnover(x['K_list'], x['K_list_prev']), axis=1)
        df_q = df_q[['D', 'T', self.factor_name]].rename(columns={self.factor_name: self.factor_forward})
        return df_q

    def calculate_lsreturn(self):
        """4.4 LongShort部分之计算LSreturn"""
        df_lsreturn = self.df_longshort.groupby(by=['D', 'T', 'q'], as_index=False)[self.forward_return_name].mean()
        df_lsreturn = df_lsreturn.pivot(index=['D', 'T'], columns='q', values=self.forward_return_name).reset_index()
        df_lsreturn['LSreturns'] = df_lsreturn.apply(lambda row: row['Q10'] - row['Q1'], axis=1)
        self.backtest_result['LSreturns'] = df_lsreturn[['D', 'T', 'LSreturns']].rename(columns={"LSreturns": self.factor_forward})
        self.backtest_result['top_LSreturns'] = df_lsreturn[['D', 'T', 'Q10']].rename(columns={"Q10": self.factor_forward})
        self.backtest_result['bottom_LSreturns'] = df_lsreturn[['D', 'T', 'Q1']].rename(columns={"Q1": self.factor_forward})
        del df_lsreturn

    def calculate_hitrate(self):
        """4.5 LongShort部分之计算HitRate"""

        def _agg_calc_hitrate(arrs):
            pos_neg_len = len(arrs)
            if pos_neg_len == 0:
                return [0.0, 0.0, 0.0]
            pos_count = sum(1 if (arr[1] == 'Q1' and arr[0] > 0) else 0 for arr in arrs)
            pos_len = len([arr for arr in arrs if arr[1] == 'Q1'])
            neg_count = sum(1 if (arr[1] == 'Q10' and arr[0] < 0) else 0 for arr in arrs)
            neg_len = len([arr for arr in arrs if arr[1] == 'Q10'])
            hitrate = float((pos_count + neg_count) / pos_neg_len)
            top_hitrate = 0.0 if pos_len == 0 else float(pos_count / pos_len)
            bottom_hitrate = 0.0 if neg_len == 0 else float(neg_count / neg_len)
            return [hitrate, top_hitrate, bottom_hitrate]

        df_hitrate = self.df_longshort.copy()
        df_hitrate['FQ'] = df_hitrate.apply(lambda row: [row[self.forward_return_name], row['q']], axis=1)
        df_hitrate = df_hitrate.groupby(by=['D', 'T'], as_index=False).agg(
            hitRates=('FQ', _agg_calc_hitrate),
        )
        df_hitrate['HitRate'] = df_hitrate['hitRates'].map(lambda x: x[0])
        df_hitrate['top_HitRate'] = df_hitrate['hitRates'].map(lambda x: x[1])
        df_hitrate['bottom_HitRate'] = df_hitrate['hitRates'].map(lambda x: x[2])
        self.backtest_result['HitRate'] = df_hitrate[['D', 'T', 'HitRate']].rename(
            columns={"HitRate": self.factor_forward})
        self.backtest_result['top_HitRate'] = df_hitrate[['D', 'T', 'top_HitRate']].rename(
            columns={"top_HitRate": self.factor_forward})
        self.backtest_result['bottom_HitRate'] = df_hitrate[['D', 'T', 'bottom_HitRate']].rename(
            columns={"bottom_HitRate": self.factor_forward})
        del df_hitrate

    def calculate_wealth(self):
        """4.6 Longshort部分之Wealth模块"""
        df_lsreturn = self.backtest_result['LSreturns'].copy()
        df_lsreturn[self.factor_forward] = df_lsreturn[self.factor_forward].apply(lambda x: x+1).cumprod()
        df_lsreturn_top = self.backtest_result['top_LSreturns'].copy()
        df_lsreturn_top[self.factor_forward] = df_lsreturn_top[self.factor_forward].apply(lambda x: x+1).cumprod()
        df_lsreturn_bottom = self.backtest_result['bottom_LSreturns'].copy()
        df_lsreturn_bottom[self.factor_forward] = df_lsreturn_bottom[self.factor_forward].apply(lambda x: x+1).cumprod()
        self.backtest_result['Wealth'] = df_lsreturn
        self.backtest_result['top_Wealth'] = df_lsreturn_top
        self.backtest_result['bottom_Wealth'] = df_lsreturn_bottom
        del df_lsreturn, df_lsreturn_top, df_lsreturn_bottom

    def _coverage_summary(self):
        self.coverage['coverage_max'] = self.factor_data['K'].nunique()
        total_days = self.factor_data['D'].nunique()
        df_days = self.factor_data.groupby('K')['D'].count()
        coverage_min = len(df_days[df_days == total_days])
        self.coverage['coverage_min'] = coverage_min

    @staticmethod
    def _max_drawdown(xs: list) -> float:
        mdd = 0
        maxwealth = 1
        x = iter(xs)
        while True:
            try:
                value = next(x)
            except Exception as e:
                return mdd
            maxwealth = max(maxwealth, value)
            mdd = min(mdd, value/maxwealth - 1)

    def calculate_serial_corr(self):
        temp_df = self.factor_data.sort_values(by=['D', 'T'], ascending=True)
        temp_df = temp_df[['K', 'D', 'T', self.factor_name]]
        """获取前一天的因子序列"""
        prev_fac = pd.pivot_table(temp_df, columns='K', index=["D", "T"], values=self.factor_name).shift(1)
        prev_fac = prev_fac.unstack(['D', 'T']).reset_index()
        prev_fac.columns = ['K', 'D', 'T', 'prev_fac']
        prev_fac = prev_fac.dropna(subset=['prev_fac'])
        """滚动计算相关性"""
        temp_df = pd.merge(temp_df, prev_fac, how='left', on=['K', 'D', 'T'])
        serial_correlation = temp_df.groupby(['D', 'T']).apply(
            lambda x: x[[self.factor_name, 'prev_fac']].corr(method='spearman').iloc[0, 1])
        serial_correlation = serial_correlation.to_frame().reset_index().dropna()
        serial_correlation.columns = ['D', 'T', self.factor_forward]
        self.backtest_result['SCs'] = serial_correlation
        del temp_df

    def summary(self, summary_type="top"):
        summary = dict({
            'fac_name': self.factor_name,
            'backtest_sd': self._factor_config.sd,
            'backtest_ed': self._factor_config.ed
        })
        summary['coverage_max'] = self.coverage['coverage_max']
        summary['coverage_min'] = self.coverage['coverage_min']
        df_ic = self.backtest_result['ICs'].copy()
        if summary_type.lower() == 'top':
            df_wealth = self.backtest_result['top_Wealth'].copy()
            df_lsreturn = self.backtest_result['top_LSreturns'].copy()
            df_hitrate = self.backtest_result['top_HitRate'].copy()
            df_turnover = self.backtest_result['top_HitRate'].copy()
            summary_key = 'top_Summary'
        elif summary_type.lower() == 'bottom':
            df_wealth = self.backtest_result['bottom_Wealth'].copy()
            df_lsreturn = self.backtest_result['bottom_LSreturns'].copy()
            df_hitrate = self.backtest_result['bottom_HitRate'].copy()
            df_turnover = self.backtest_result['bottom_HitRate'].copy()
            summary_key = 'bottom_Summary'
        else:
            df_wealth = self.backtest_result['Wealth'].copy()
            df_lsreturn = self.backtest_result['LSreturns'].copy()
            df_ic = self.backtest_result['ICs'].copy()
            df_hitrate = self.backtest_result['HitRate'].copy()
            df_turnover = self.backtest_result['HitRate'].copy()
            summary_key = 'Summary'
        n_days = df_wealth.shape[0]
        btm_w = df_wealth.iloc[0, df_wealth.columns.get_loc(self.factor_forward)]
        btm_d1 = 1.0 if n_days < abs(LAST_ONE_DAY) else df_wealth.iloc[LAST_ONE_DAY, df_wealth.columns.get_loc(self.factor_forward)]
        btm_m1 = 1.0 if n_days < abs(LAST_ONE_MONTH) else df_wealth.iloc[LAST_ONE_MONTH, df_wealth.columns.get_loc(self.factor_forward)]
        btm_m3 = 1.0 if n_days < abs(LAST_THREE_MONTH) else df_wealth.iloc[LAST_THREE_MONTH, df_wealth.columns.get_loc(self.factor_forward)]
        btm_y1 = 1.0 if n_days < abs(LAST_ONE_YEAR) else df_wealth.iloc[LAST_ONE_YEAR, df_wealth.columns.get_loc(self.factor_forward)]
        btm_y3 = 1.0 if n_days < abs(LAST_THREE_YEAR) else df_wealth.iloc[LAST_THREE_YEAR, df_wealth.columns.get_loc(self.factor_forward)]
        btm_y5 = 1.0 if n_days < abs(LAST_FIVE_YEAR) else df_wealth.iloc[LAST_FIVE_YEAR, df_wealth.columns.get_loc(self.factor_forward)]
        return_z = btm_d1/btm_w - 1
        summary['return_1m'] = btm_d1/btm_m1 - 1
        summary['return_3m'] = btm_d1/btm_m3 - 1
        return_z1y = btm_d1/btm_y1 - 1
        return_z3y = btm_d1/btm_y3 - 1
        return_z5y = btm_d1/btm_y5 - 1
        summary['cagr'] = (return_z + 1) ** (DAYS_PER_YEAR / n_days) - 1
        summary['cagr_1y'] = summary['cagr'] if n_days < abs(LAST_ONE_YEAR) else return_z1y
        summary['cagr_3y'] = summary['cagr'] if n_days < abs(LAST_THREE_YEAR) else (
                (return_z3y + 1) ** (DAYS_PER_YEAR / abs(LAST_THREE_YEAR)) - 1.0)
        summary['cagr_5y'] = summary['cagr'] if n_days < abs(LAST_FIVE_YEAR) else (
                (return_z5y + 1) ** (DAYS_PER_YEAR / abs(LAST_FIVE_YEAR)) - 1.0)
        summary['volatility'] = df_lsreturn[self.factor_forward].std(ddof=1) * np.sqrt(DAYS_PER_YEAR)
        summary['sharpe_ratio'] = summary['cagr'] / summary['volatility']
        summary['rank_ic_mean'] = df_ic[self.factor_forward].mean()
        summary['rank_ic_std'] = df_ic[self.factor_forward].std(ddof=1)
        summary['ic_ir'] = summary['rank_ic_mean'] / summary['rank_ic_std']
        summary['hit_rate'] = df_hitrate[self.factor_forward].mean() * 100.0
        summary['max_drawdown'] = self._max_drawdown(df_wealth[self.factor_forward].tolist())
        summary['serial_corr'] = self.backtest_result['SCs'][self.factor_forward].mean()
        summary['turnover'] = df_turnover[self.factor_forward].mean() * 100
        cagr_res = self._cagr_summary(df_wealth)
        rank_ic_mean_res = self._rank_ic_mean_summary(df_ic)
        summary.update(cagr_res)
        summary.update(rank_ic_mean_res)
        self.backtest_result[summary_key] = summary
        del df_wealth, df_lsreturn, df_hitrate, df_turnover, df_ic

    def _cagr_summary(self, df_wealth):
        """
        CAGR算法
        """
        cagr_res = {}
        df_wealth['Y'] = df_wealth['D'].apply(lambda day: str(day)[0:4])
        df_cagr_agg = df_wealth.groupby(by=['Y'], as_index=False).agg(M=('D', 'count'), D_max=('D', 'max'))
        df_cagr_agg = pd.merge(df_cagr_agg, df_wealth[['D', self.factor_forward]], left_on='D_max', right_on='D')
        df_cagr_agg = df_cagr_agg.rename(columns={self.factor_forward: "D_max_wealth"})
        df_cagr_agg.drop('D', axis=1, inplace=True)
        df_cagr_agg['D_max_wealth_1'] = df_cagr_agg['D_max_wealth'].shift(1)
        for _, row in df_cagr_agg.iterrows():
            cagr_year_key = "cagr_year_{}".format(row['Y'])
            if str(row['D_max_wealth_1']) == 'nan':
                cagr_res[cagr_year_key] = row['D_max_wealth'] ** (
                        row['M'] / float(DAYS_PER_YEAR)) - 1.0
            else:
                cagr_res[cagr_year_key] = (row['D_max_wealth'] / row['D_max_wealth_1']) ** (
                        row['M'] / float(DAYS_PER_YEAR)) - 1.0
        return cagr_res

    def _rank_ic_mean_summary(self, df_ic):
        """
        计算 Rank IC Mean Summary
        :param df_ic:
        :return:
        """
        rank_ic_mean_res = {}
        df_ic['Y'] = df_ic['D'].apply(lambda day: str(day)[0:4])
        df_ic_agg = df_ic.groupby(by=['Y'], as_index=False).agg(rank_ic_mean_year=(self.factor_forward, 'mean'))
        for _, row in df_ic_agg.iterrows():
            rank_ic_mean_res["rank_ic_mean_year_{}".format(row['Y'])] = row['rank_ic_mean_year']
        return rank_ic_mean_res


class Factor:
    """Factor类,表示单个因子"""

    def __init__(self, factor_name, factor_path):
        self.factor_name = factor_name
        self.factor_path = factor_path
        self.files_shouldbe = self._load_factor_files(factor_path)
        self.data = None

    def _load_factor_files(self, path: str) -> list:  # noqa
        alldays = FatbullsFactorContext.get_instance().get_alldays()
        csvdays = [file.split(".")[0] for file in os.listdir(path) if file.endswith(".csv")]
        if len(csvdays) == 0:
            raise FileNotFoundError("no csv file found in path : {}".format(path))
        minday = min(csvdays)
        files_shouldbe = list()
        missing_files = list()
        for day in alldays:
            if day < minday:
                continue
            csv_file = os.path.join(path, "{}.csv".format(day))
            if os.path.isfile(csv_file):
                files_shouldbe.append(csv_file)
            else:
                missing_files.append(csv_file)
        if len(missing_files) > 0:
            raise FileNotFoundError("the following files are missing: {}".format(",".join(missing_files)))
        return files_shouldbe

    def load_data(self, use_cache=True):
        # 从数据库加载投资范围数据
        try:
            if use_cache and (self.data is not None) and (not self.data):
                return self.data
            self.data = {}
            cnall = FatbullsFactorContext.get_instance().get_univ_cnall()
            for filepath in self.files_shouldbe:
                dtcsv = pd.read_csv(filepath, header=0, sep=",")
                dtcsv["D"] = dtcsv["D"].astype(str)
                self._checkformat(filepath, dtcsv, cnall.data)  # 查一下文件结构是否符合规范
                self.data[dtcsv["D"].iloc[0]] = dtcsv  # 确保一个文件里只有一个D，已经查过了
        except Exception as e:
            raise FactorBackTestException(
                'load factor data error, factor_name:{}, msg: {}'.format(self.factor_name, str(e)))
        return self.data

    def _checkformat(self, filepath: str, df: pd.DataFrame, cnall):
        """
        查每个文件里的格式
        """
        # 查表头
        if not {"K", "D", "T", self.factor_name} == set(df.columns):
            raise KeyError("incorrect header format in file {}".format(filepath))

        # 查文件内是否有重复stock code
        df_dd = df.drop_duplicates(subset=["K"])
        if len(df_dd) < len(df):
            raise ValueError("duplicate stock codes in file {}".format(filepath))

        # 查文件内日期是否与文件名一致，是否有不同日期出现在一个文件
        dt = (filepath.split("/")[-1]).split(".")[0]
        try:
            assert max(df.D) == min(df.D) == dt
        except AssertionError:
            raise ValueError(
                "multiple dates in same file/filename doesn't match D column in file {}".format(filepath)
            )

        """ 查文件内的股票代码是否都是上市的股票 """
        stk_code = set(df_dd["K"])
        if not set(stk_code) <= set(cnall[dt]["K"]):
            raise ValueError("stock code beyond cnall exists in file {}".format(filepath))


class Universe:
    """Universe类,表示投资范围"""

    def __init__(self, name, start_date, end_date):
        self.name = name
        self.start_date = start_date
        self.end_date = end_date
        self._data = None

    def _partition_df(self, df: pd.DataFrame, col: str):  # noqa
        """Partition a DataFrame into a dictionary of DataFrames by column value."""
        return {k: v.reset_index(drop=True) for k, v in df.groupby(col, sort=False)}

    def load_data(self):
        # 从数据库加载投资范围数据
        try:
            univ = fqdatac.get_universe(sd=self.start_date, ed=self.end_date, univ_name=self.name)
            self._data = self._partition_df(univ, "D")
        except Exception as e:
            raise FactorBackTestException('load universe data error, universe:{}, msg: {}'.format(self.name, str(e)))

    @property
    def data(self):
        if self._data is None:
            self.load_data()
        return self._data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: fac_backtest.py
Author: wuyiyang(wuyiyang@hcyjs.com), ....
Date: 2023/8/28 10:54:43
"""

import logging
from . import  fac_class as fac


def backtest_main(
        backtest_name: str,
        sd: str,
        ed: str,
        task_num: str,
        univ_name: str = "cnall",
        fdays: int = 1,
        rettype: str = "vwap30 to vwap30",
        directory_list: list = [],
        output_path: str = ''
):
    # 实例化
    loader = fac.FactorBTLoader(
        sd=sd,
        ed=ed,
        backtest_name=backtest_name,
        univ_name=univ_name,
        fdays=fdays,
        rettype=rettype,
        directory_list=directory_list,
    )

    # 数据加载与预处理
    logging.info("start factor loading")
    for d in directory_list:
        loader.load(
            directory=d,
        )
        logging.info(f"factor {d.split('/')[-1]} successfully loaded")
    loader.onlyuniv()

    logging.info("start fwdret loading")
    loader.fwdret_load()

    # 整个回测过程
    logging.info("start factor evaluation")
    evaluator = fac.FactorEvaluator(
        sd=loader.sd,
        ed=loader.ed,
        backtest_name=loader.backtest_name,
        univ_name=loader.univ_name,
        fdays=loader.fdays,
        rettype=loader.rettype,
        directory_list=loader.directory_list,
        raw=loader.container,
        fwdret=loader.fwdret1,
        output_path=output_path,
        task_num=task_num,
    )
    evaluator.initialize()
    evaluator.nextfactor()
    while evaluator.whole_processing:
        # evaluator.nextfile()
        # while evaluator.fac_processing:
        #     evaluator.evaluate()
        #     evaluator.save_lastfile()
        #     evaluator.nextfile()
        logging.info(f"factor evaluation for factor{evaluator.curfac}")
        evaluator.evaluate()
        evaluator.fac_report()
        evaluator.nextfactor()
        logging.info("--next factor")
    evaluator.output()


def run(
        backtest_name: str,
        sd: str,
        ed: str,
        task_num: str = "b",
        univ_name: str = "cnall",
        fdays: int = 1,
        rettype: str = "vwap30",
        directory_list: list = [],
        output_path: str = '/data/quant/report/factor_backtest'
):
    fac.FQDataLoader()
    backtest_main(
        backtest_name=backtest_name,
        sd=sd,
        ed=ed,
        task_num=task_num,
        univ_name=univ_name,
        fdays=fdays,
        rettype=rettype,
        directory_list=directory_list,
        output_path=output_path
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: run_single_calculate.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/21 18:50:29
"""


import pandas as pd
import numpy as np
from fatbulls import time_cost

"""基础函数区域"""

FIRST_DAY = 0
LAST_ONE_DAY = -1
LAST_ONE_MONTH = -22
LAST_THREE_MONTH = -64
LAST_ONE_YEAR = -245
LAST_THREE_YEAR = -490
LAST_FIVE_YEAR = -1225
DAYS_PER_YEAR = 244


class FactorCalculateTask(object):

    def __init__(self, factor_name, forward_return_name, forward_days, return_type):
        self.factor_name = factor_name
        self.forward_return_name = forward_return_name
        self.factor_data = pd.read_csv('./output/{}_merged.csv'.format(factor_name), sep=",", header='infer')
        self.backtest_result = dict()
        self.output_colname = "{}_{}".format(factor_name, forward_return_name)
        self.coverage = {}

    def _middle_calculate(self):
        self.factor_data = self.factor_data.sort_values(by=['D', 'T', 'K'], axis=0, ascending=True)
        self.grouped = self.factor_data.groupby(by=['D', 'T'], as_index=False)
        self.factor_data['q'] = self.factor_data.groupby(by=['D', 'T'], as_index=False)[self.factor_name].transform(
            lambda x: pd.qcut(x, 10, labels=["Q{}".format(i) for i in range(1, 11)]))
        self.factor_data['q'] = self.factor_data['q'].astype(str)
        self.df_longshort = self.factor_data.query("q == 'Q1' or q == 'Q10'")

    @time_cost
    def run(self, *args, **kwargs):
        self._middle_calculate()
        self._coverage_summary()
        self.calculate_ic()
        self.calculate_coverage()
        self.calculate_turnover()
        self.calculate_lsreturn()
        self.calculate_hitrate()
        self.calculate_wealth()
        self.summary()
        self.summary(summary_type="top")
        self.summary(summary_type="bottom")
        return self.backtest_result

    def calculate_ic(self):
        """4.1 计算IC相关性, 不区分LongShort"""
        df_ic = self.grouped.apply(lambda x: x[self.factor_name].corr(x[self.forward_return_name], 'spearman'))
        df_ic.columns = ['D', 'T', self.output_colname]
        self.backtest_result['ICs'] = df_ic
        del df_ic

    def calculate_coverage(self):
        """4.1 计算coverage, 不区分LongShort"""
        df_coverage = self.grouped.agg({'K': 'count'})
        df_coverage = df_coverage.rename(columns={"K": self.factor_name})
        df_coverage = df_coverage[['D', 'T', self.factor_name]].rename(columns={self.factor_name: self.output_colname})
        self.backtest_result['Coverage'] = df_coverage
        del df_coverage

    def calculate_turnover(self):
        """
            4.1 urnover , 分为LongShort、Long、Short
            LongShort Turnover = 1 - (C1 + C10)/(S1 + S10)
            Long Turnover  = 1 - C10/S10
            Short Turnover = 1 - C1/S1
        """
        self.backtest_result['Turnover'] = self._calculate_turnover(self.df_longshort)
        self.backtest_result['top_Turnover'] = self._calculate_turnover(self.df_longshort.query("q == 'Q1'"))
        self.backtest_result['bottom_Turnover'] = self._calculate_turnover(self.df_longshort.query("q == 'Q10'"))

    def _calculate_turnover(self, df_q: pd.DataFrame):
        def _calc_turnover(x, y):
            """计算交集"""
            if x is None or y is None:
                return 1.0
            return 1.0 - len(set(x).intersection(set(y))) / (len(x) + len(y))

        df_q = df_q.groupby(['D', 'T'], as_index=False).agg({"K": list})
        df_q.columns = ['D', 'T', 'K_list']
        df_q['K_list_prev'] = df_q.sort_values(by=['D', 'T'], ascending=True)['K_list'].shift(1)
        df_q[self.factor_name] = df_q.apply(lambda x: _calc_turnover(x['K_list'], x['K_list_prev']), axis=1)
        df_q = df_q[['D', 'T', self.factor_name]].rename(columns={self.factor_name: self.output_colname})
        return df_q

    def calculate_lsreturn(self):
        """4.4 LongShort部分之计算LSreturn"""
        df_lsreturn = self.df_longshort.groupby(by=['D', 'T', 'q'], as_index=False)[self.forward_return_name].mean()
        df_lsreturn = df_lsreturn.pivot(index=['D', 'T'], columns='q', values=self.forward_return_name).reset_index()
        df_lsreturn['LSreturns'] = df_lsreturn.apply(lambda row: row['Q10'] - row['Q1'], axis=1)
        self.backtest_result['LSreturns'] = df_lsreturn[['D', 'T', 'LSreturns']].rename(columns={"LSreturns": self.output_colname})
        self.backtest_result['top_Returns'] = df_lsreturn[['D', 'T', 'Q10']].rename(columns={"Q10": self.output_colname})
        self.backtest_result['bottom_Returns'] = df_lsreturn[['D', 'T', 'Q1']].rename(columns={"Q1": self.output_colname})
        del df_lsreturn

    def calculate_hitrate(self):
        """4.5 LongShort部分之计算HitRate"""

        def _agg_calc_hitrate(arrs):
            pos_neg_len = len(arrs)
            if pos_neg_len == 0:
                return [0.0, 0.0, 0.0]
            pos_count = sum(1 if (arr[1] == 'Q1' and arr[0] > 0) else 0 for arr in arrs)
            pos_len = len([arr for arr in arrs if arr[1] == 'Q1'])
            neg_count = sum(1 if (arr[1] == 'Q10' and arr[0] < 0) else 0 for arr in arrs)
            neg_len = len([arr for arr in arrs if arr[1] == 'Q10'])
            hitrate = float((pos_count + neg_count) / pos_neg_len)
            top_hitrate = 0.0 if pos_len == 0 else float(pos_count / pos_len)
            bottom_hitrate = 0.0 if neg_len == 0 else float(neg_count / neg_len)
            return [hitrate, top_hitrate, bottom_hitrate]

        df_hitrate = self.df_longshort.copy()
        df_hitrate['FQ'] = df_hitrate.apply(lambda row: [row[self.forward_return_name], row['q']], axis=1)
        df_hitrate = df_hitrate.groupby(by=['D', 'T'], as_index=False).agg(
            hitRates=('FQ', _agg_calc_hitrate),
        )
        df_hitrate['HitRate'] = df_hitrate['hitRates'].map(lambda x: x[0])
        df_hitrate['top_HitRate'] = df_hitrate['hitRates'].map(lambda x: x[1])
        df_hitrate['bottom_HitRate'] = df_hitrate['hitRates'].map(lambda x: x[2])
        self.backtest_result['HitRate'] = df_hitrate[['D', 'T', 'HitRate']].rename(
            columns={"HitRate": self.output_colname})
        self.backtest_result['top_HitRate'] = df_hitrate[['D', 'T', 'top_HitRate']].rename(
            columns={"top_HitRate": self.output_colname})
        self.backtest_result['bottom_HitRate'] = df_hitrate[['D', 'T', 'bottom_HitRate']].rename(
            columns={"bottom_HitRate": self.output_colname})
        del df_hitrate

    def calculate_wealth(self):
        """4.6 Longshort部分之Wealth模块"""
        df_lsreturn = self.backtest_result['LSreturns'].copy()
        df_lsreturn[self.output_colname] = df_lsreturn[self.output_colname].apply(lambda x: x+1).cumprod()
        df_lsreturn_top = self.backtest_result['top_Returns'].copy()
        df_lsreturn_top[self.output_colname] = df_lsreturn_top[self.output_colname].apply(lambda x: x+1).cumprod()
        df_lsreturn_bottom = self.backtest_result['bottom_Returns'].copy()
        df_lsreturn_bottom[self.output_colname] = df_lsreturn_bottom[self.output_colname].apply(lambda x: x+1).cumprod()
        self.backtest_result['Wealth'] = df_lsreturn
        self.backtest_result['top_Wealth'] = df_lsreturn_top
        self.backtest_result['bottom_Wealth'] = df_lsreturn_bottom
        del df_lsreturn, df_lsreturn_top, df_lsreturn_bottom

    def _coverage_summary(self):
        self.coverage['coverage_max'] = self.factor_data['K'].nunique()
        total_days = self.factor_data['D'].nunique()
        df_days = self.factor_data.groupby('K')['D'].count()
        coverage_min = len(df_days[df_days == total_days])
        self.coverage['coverage_min'] = coverage_min

    @staticmethod
    def _max_drawdown(xs: list) -> float:
        mdd = 0
        maxwealth = 1
        x = iter(xs)
        while True:
            try:
                value = next(x)
            except Exception as e:
                return mdd
            maxwealth = max(maxwealth, value)
            mdd = min(mdd, value/maxwealth - 1)

    def _calculate_serial_corr(self):
        temp_df = self.factor_data.sort_values(by='D', ascending=True)
        temp_df = temp_df[['K', 'D', self.factor_name]]
        """获取前一天的因子序列"""
        prev_fac = pd.pivot_table(temp_df, columns='K', index="D", values='DTech_amt').shift(1)
        prev_fac = prev_fac.unstack().reset_index()
        prev_fac.columns = ['K', 'D', 'prev_fac']
        prev_fac = prev_fac.dropna(subset=['prev_fac'])
        """滚动计算相关性"""
        temp_df = pd.merge(temp_df, prev_fac, how='left', on=['K', 'D'])
        serial_correlation = temp_df.groupby('D').apply(
            lambda x: x[['DTech_amt', 'prev_fac']].corr(method='spearman').iloc[0, 1])
        serial_correlation = serial_correlation.to_frame().reset_index().dropna()
        serial_correlation.columns = ['D', 'DTech_amt']
        del temp_df
        return serial_correlation

    def summary(self, summary_type="top"):
        summary = dict()
        summary['coverage_max'] = self.coverage['coverage_max']
        summary['coverage_min'] = self.coverage['coverage_min']
        df_ic = self.backtest_result['ICs'].copy()
        if summary_type.lower() == 'top':
            df_wealth = self.backtest_result['top_Wealth'].copy()
            df_lsreturn = self.backtest_result['top_Returns'].copy()
            df_hitrate = self.backtest_result['top_HitRate'].copy()
            df_turnover = self.backtest_result['top_HitRate'].copy()
            summary_key = 'top_Summary'
        elif summary_type.lower() == 'bottom':
            df_wealth = self.backtest_result['bottom_Wealth'].copy()
            df_lsreturn = self.backtest_result['bottom_Returns'].copy()
            df_hitrate = self.backtest_result['bottom_HitRate'].copy()
            df_turnover = self.backtest_result['bottom_HitRate'].copy()
            summary_key = 'bottom_Summary'
        else:
            df_wealth = self.backtest_result['Wealth'].copy()
            df_lsreturn = self.backtest_result['LSreturns'].copy()
            df_ic = self.backtest_result['ICs'].copy()
            df_hitrate = self.backtest_result['HitRate'].copy()
            df_turnover = self.backtest_result['HitRate'].copy()
            summary_key = 'Summary'
        n_days = df_wealth.shape[0]
        btm_w = df_wealth.iloc[0, df_wealth.columns.get_loc(self.output_colname)]
        btm_d1 = 1.0 if n_days < abs(LAST_ONE_DAY) else df_wealth.iloc[LAST_ONE_DAY, df_wealth.columns.get_loc(self.output_colname)]
        btm_m1 = 1.0 if n_days < abs(LAST_ONE_MONTH) else df_wealth.iloc[LAST_ONE_MONTH, df_wealth.columns.get_loc(self.output_colname)]
        btm_m3 = 1.0 if n_days < abs(LAST_THREE_MONTH) else df_wealth.iloc[LAST_THREE_MONTH, df_wealth.columns.get_loc(self.output_colname)]
        btm_y1 = 1.0 if n_days < abs(LAST_ONE_YEAR) else df_wealth.iloc[LAST_ONE_YEAR, df_wealth.columns.get_loc(self.output_colname)]
        btm_y3 = 1.0 if n_days < abs(LAST_THREE_YEAR) else df_wealth.iloc[LAST_THREE_YEAR, df_wealth.columns.get_loc(self.output_colname)]
        btm_y5 = 1.0 if n_days < abs(LAST_FIVE_YEAR) else df_wealth.iloc[LAST_FIVE_YEAR, df_wealth.columns.get_loc(self.output_colname)]
        return_z = btm_d1/btm_w - 1
        summary['return_1m'] = btm_d1/btm_m1 - 1
        summary['return_3m'] = btm_d1/btm_m3 - 1
        return_z1y = btm_d1/btm_y1 - 1
        return_z3y = btm_d1/btm_y3 - 1
        return_z5y = btm_d1/btm_y5 - 1
        summary['cagr'] = (return_z + 1) ** (DAYS_PER_YEAR / n_days) - 1
        summary['cagr_1y'] = summary['cagr'] if n_days < abs(LAST_ONE_YEAR) else return_z1y
        summary['cagr_3y'] = summary['cagr'] if n_days < abs(LAST_THREE_YEAR) else (
                    (return_z3y + 1) ** (DAYS_PER_YEAR / abs(LAST_THREE_YEAR)) - 1.0)
        summary['cagr_5y'] = summary['cagr'] if n_days < abs(LAST_FIVE_YEAR) else (
                    (return_z5y + 1) ** (DAYS_PER_YEAR / abs(LAST_FIVE_YEAR)) - 1.0)
        summary['volatility'] = df_lsreturn[self.output_colname].std(ddof=1) * np.sqrt(DAYS_PER_YEAR)
        summary['sharpe_ratio'] = summary['cagr'] / summary['volatility']
        summary['rank_ic_mean'] = df_ic[self.output_colname].mean()
        summary['rank_ic_std'] = df_ic[self.output_colname].std(ddof=1)
        summary['ic_ir'] = summary['rank_ic_mean'] / summary['rank_ic_std']
        summary['hit_rate'] = df_hitrate[self.output_colname].mean() * 100.0
        summary['max_drawdown'] = self._max_drawdown(df_wealth[self.output_colname].tolist())
        summary['serial_corr'] = self._calculate_serial_corr()
        summary['turnover'] = df_turnover[self.output_colname].mean() * 100
        cagr_res = self._cagr_summary(df_wealth)
        rank_ic_mean_res = self._rank_ic_mean_summary(df_ic)
        summary.update(cagr_res)
        summary.update(rank_ic_mean_res)
        self.backtest_result[summary_key] = summary
        del df_wealth, df_lsreturn, df_hitrate, df_turnover, df_ic

    def _cagr_summary(self, df_wealth):
        """
        CAGR算法
        """
        cagr_res = {}
        df_wealth['Y'] = df_wealth['D'].apply(lambda day: str(day)[0:4])
        df_cagr_agg = df_wealth.groupby(by=['Y'], as_index=False).agg(M=('D', 'count'), D_max=('D', 'max'))
        df_cagr_agg = pd.merge(df_cagr_agg, df_wealth[['D', self.output_colname]], left_on='D_max', right_on='D')
        df_cagr_agg = df_cagr_agg.rename(columns={self.output_colname: "D_max_wealth"})
        df_cagr_agg.drop('D', axis=1, inplace=True)
        df_cagr_agg['D_max_wealth_1'] = df_cagr_agg['D_max_wealth'].shift(1)
        for _, row in df_cagr_agg.iterrows():
            cagr_year_key = "cagr_year_{}".format(row['Y'])
            if str(row['D_max_wealth_1']) == 'nan':
                cagr_res[cagr_year_key] = row['D_max_wealth'] ** (
                            row['M'] / float(DAYS_PER_YEAR)) - 1.0
            else:
                cagr_res[cagr_year_key] = (row['D_max_wealth'] / row['D_max_wealth_1']) ** (
                            row['M'] / float(DAYS_PER_YEAR)) - 1.0
        return cagr_res

    def _rank_ic_mean_summary(self, df_ic):
        """
        计算 Rank IC Mean Summary
        :param df_ic:
        :return:
        """
        rank_ic_mean_res = {}
        df_ic['Y'] = df_ic['D'].apply(lambda day: str(day)[0:4])
        df_ic_agg = df_ic.groupby(by=['Y'], as_index=False).agg(rank_ic_mean_year=(self.output_colname, 'mean'))
        for _, row in df_ic_agg.iterrows():
            rank_ic_mean_res["rank_ic_mean_year_{}".format(row['Y'])] = row['rank_ic_mean_year']
        return rank_ic_mean_res


a = FactorCalculateTask('DTech_amt', 'fwd1_vwap30', 1, 'vwap30')
print(a.run())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: __init__.py.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 16:09:11
"""
import time
import logging
import os
from functools import wraps

__all__ = [
    '__version__', 'run_factor_backtest', 'run_portfolio_backtest', 'time_cost'
]


def time_cost(func):
    @wraps(func)
    def wrap(*args, **kwargs):
        begin_time = time.perf_counter()
        result = func(*args, **kwargs)
        start_time = time.perf_counter()
        tips = '函数: %r 耗时: %2.4f 秒' % (func.__name__, start_time - begin_time)
        logging.info(tips)
        print(tips)
        return result
    return wrap


@time_cost
def run_factor_backtest(**kwargs):
    """
    传入约定函数和因子配置运行回测。约定函数详见 API 手册约定函数部分，可用的配置项详见参数配置部分。

    :Keyword Arguments:
        * **config** (dict) -- 策略配置字典

    :return: dict

    """
    from rqalpha.utils.functools import clear_all_cached_functions
    from rqalpha.utils.config import parse_config
    from fatbulls import main

    config = kwargs.get('config', None)
    if config is None:
        config = {}
    else:
        if isinstance(config, dict):
            try:
                del config["base"]["strategy_file"]
                logging.warning("delete strategy_file successfully!")
            except KeyError as e:
                logging.info("base strategy_file not found")
    fatbulls_config_path = kwargs.get('config_path', None)
    if fatbulls_config_path is None:
        fatbulls_config_path = os.path.join(os.path.dirname(__file__), 'resource/fatbulls_config.yml')
    config = parse_config(config, config_path=fatbulls_config_path)
    clear_all_cached_functions()

    return main.run(config)


@time_cost
def run_portfolio_backtest(**kwargs):
    """
    投资组合回测

    :Keyword Arguments:
        * **config** (dict) -- 组合回测配置字典

    :return: dict

    """

    pass


__version__ = '0.0.1'




#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: read_backtest_config.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/16 15:50:03
"""

from fatbulls.models.factor_model import BackTestConfig


if __name__ == "__main__":
    backtest_config = BackTestConfig('20230101', '20230701', factor_dirs=[
        '/data/quant/factors/DTech/DTech_volume',
        '/data/quant/factors/DTech/DTech_amt'])
    print(backtest_config.today)
    print(backtest_config.sd)
    print(backtest_config.ed)
    print(backtest_config.factor_dirs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: context.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 16:33:27
"""
import logging

from rqalpha.core.global_var import GlobalVars
from .models.factor_model import BackTestConfig


class FatbullsFactorContext(object):
    _fbctx = None  # type: FatbullsFactorContext

    def __init__(self, config):
        FatbullsFactorContext._fbctx = self
        self.config = config
        self.global_vars = GlobalVars()
        self.alldays = None
        self.univ_cnall = None
        self.univ_userselect = None
        self.backtest_config = None

    @classmethod
    def get_instance(cls):
        """
        返回已经创建的 Context 对象
        """
        if FatbullsFactorContext._fbctx is None:
            raise RuntimeError(
                "FatbullsContext has not been created."
                " Please Use `FatbullsContext.get_instance()` after FatBulls init")
        return FatbullsFactorContext._fbctx

    def set_config(self, config):
        self.config = config

    def get_config(self):
        return self.config

    def get_factor_config(self) -> BackTestConfig:
        try:
            if self.backtest_config is not None:
                return self.backtest_config
            self.backtest_config = BackTestConfig(
                sd=self.config.factor.sd,
                ed=self.config.factor.ed,
                task_num=self.config.factor.task_num,
                backtest_name=self.config.factor.backtest_name,
                univ_name=self.config.factor.univ_name,
                fdays=self.config.factor.fdays,
                rettype=self.config.factor.rettype,
                factor_dirs=self.config.factor.factor_dirs
            )
        except Exception as e:
            raise EnvironmentError('valid factor backtest config!, msg: {}'.format(str(e)))
        return self.backtest_config

    def set_alldays(self, alldays):
        self.alldays = alldays

    def get_alldays(self):
        return self.alldays

    def set_univ_cnall(self, univ_cnall):
        self.univ_cnall = univ_cnall

    def get_univ_cnall(self):
        return self.univ_cnall

    def set_univ_userselect(self, univ_userselect):
        self.univ_userselect = univ_userselect

    def get_univ_userselect(self):
        return self.univ_userselect

    def get_base_logpath(self):
        try:
            return self.config.factor.base_log_path
        except Exception as e:
            logging.warning("factor.base_log_path not exists")
            return None

    def get_base_report_dirpath(self):
        try:
            return self.config.factor.base_report_dirpath
        except Exception as e:
            raise FileNotFoundError("factor.base_report_dirpath not exists")




import numpy as np
import pandas as pd

# df = pd.read_csv('./output/DTech_amt_merged.csv', sep=",", header='infer')

#
# coverage_max = len(df['K'].unique())
# total_days = len(df['D'].unique())
# df_days = df.groupby('K')['D'].count()
# coverage_min = len(df_days[df_days == total_days])
# df_coverage = pd.DataFrame({'coverage_max': [coverage_max], 'coverage_min': [coverage_min]})
# print(df_coverage)

#
# """基础部分之分10组"""
# grouped = df.groupby(by=['D', 'T'], as_index=False)
# df['q'] = df.groupby(by=['D', 'T'], as_index=False)[factor_name].transform(
#     lambda x: pd.qcut(x, 10, labels=["Q{}".format(i) for i in range(1, 11)]))
# df['q'] = df['q'].astype(str)
# df_q = df.query("q == 'Q1' or q == 'Q10'")
# # df_q = df[df.q.isin(['Q1', 'Q10'])]


"""计算SerialCorr"""

"""
@timer
def serial_corr():
    global df
    temp_df = df.sort_values(by='D', ascending=True)
    temp_df = df[['K', 'D', 'DTech_amt']]
    #获取前一天的因子序列
    prev_fac = pd.pivot_table(temp_df, columns='K', index="D", values='DTech_amt').shift(1)
    prev_fac = prev_fac.unstack().reset_index()
    prev_fac.columns = ['K', 'D', 'prev_fac']
    prev_fac = prev_fac.dropna(subset=['prev_fac'])
    #滚动计算相关性
    temp_df = pd.merge(temp_df, prev_fac, how='left', on=['K', 'D'])
    print(temp_df)
    serial_correlation = temp_df.groupby('D').apply(
        lambda x: x[['DTech_amt', 'prev_fac']].corr(method='spearman').iloc[0, 1])
    serial_correlation = serial_correlation.to_frame().reset_index().dropna()
    serial_correlation.columns = ['D', 'DTech_amt']
    return serial_correlation


s = time.time()
df_serial_corr = serial_corr()
print(np.mean(df_serial_corr['DTech_amt']))
e = time.time()
print(e - s)

# s = time.time()
# df['DT'] = df.apply(lambda row: "{} {}".format(row['D'], row['T']), axis=1)
# df_serial_klist = df[['K', 'DT', 'DTech_amt']].pivot(index=['K'], columns='DT', values='DTech_amt').reset_index()
# cols = df_serial_klist.columns.tolist()
# cols.remove('K')
# cols = list(zip(cols, cols[1:]))
# serial_corrs = []
# for col in cols:
#     corr = df_serial_klist[col[0]].corr(df_serial_klist[col[1]], method='pearson')
#     serial_corrs.append(corr)
# serial_corr = np.mean(serial_corrs)
# print(serial_corr)
# e = time.time()
# print(e - s)
"""

# df_serial_klist = grouped.agg(k_list=('K', list))
# df_serial_klist['k_list_prev_day'] = df_serial_klist['k_list'].shift(1)
# df_serial_klist['k_intersection'] = df_serial_klist.apply(lambda x: np.nan if x['k_list_prev_day'] is None else list(set(x['k_list']) & set(x['k_list_prev_day'])), axis=1)
# df_serial_klist = df_serial_klist.dropna(subset=['k_intersection'], axis=0)
# print(df_serial_klist[['D', 'T', 'k_intersection']])
# print(df_serial_klist.to_csv('./output/x.csv'))

#
# """4.1 计算IC相关性"""
# df_ic = grouped.apply(lambda x: x['DTech_high'].corr(x[fwtret], 'spearman'))
# df_ic.columns = ['D', 'T', col_name]
# # df_ic.to_csv("ic.csv")
# # print(df_ic)
#
# """4.2 LongShort部分之计算Coverage"""
# df_coverage = grouped.agg({'K': 'count'})
# df_coverage = df_coverage.rename(columns={"K": col_name})
# df_coverage = df_coverage[['D', 'T', col_name]]
# # df_coverage.to_csv("coverage.csv")
# # print(df_coverage)
#
# """4.3 LongShort部分之计算turnover"""
# df_turnover = df_q.groupby(['D', 'T'], as_index=False).agg({"K": list})
# df_turnover.columns = ['D', 'T', 'K_list']
# df_turnover['K_list_prev'] = df_turnover.sort_values(by=['D', 'T'], ascending=True)['K_list'].shift(1)
# df_turnover[col_name] = df_turnover.apply(lambda x: calc_turnover(x['K_list'], x['K_list_prev']), axis=1)
# df_turnover = df_turnover[['D', 'T', col_name]]
# # df_turnover.to_csv("turnover.csv")
# # print(df_turnover)
#
# """4.4 LongShort部分之计算LSreturn"""
# df_lsreturn = df_q.groupby(by=['D', 'T', 'q'], as_index=False)[fwtret].mean()
# df_lsreturn = df_lsreturn.pivot(index=['D', 'T'], columns='q', values=fwtret).reset_index()
# df_lsreturn['LSreturn'] = df_lsreturn.apply(lambda row: row['Q10'] - row['Q1'], axis=1)
# df_lsreturn = df_lsreturn[['D', 'T', 'LSreturn']]
# # print(df_lsreturn)
#
# """4.5 LongShort部分之计算HitRate"""
# df_hitrate = df_q.groupby(by=['D', 'T'], as_index=False).apply(calc_hitrate)
# df_hitrate.columns = ['D', 'T', col_name]
# # df_hitrate.to_csv("hitrate.csv")
# # print(df_hitrate)
#
# """4.6 Longshort部分之Wealth模块"""
# df_lsreturn['wealth'] = df_lsreturn['LSreturn'].apply(lambda x: x+1).cumprod()
# # print(df_lsreturn)
#
# df_wealth = df_lsreturn[['D', 'T', 'wealth']]
# df_wealth.columns = ['D', 'T', col_name]
#
# df_lsreturn = df_lsreturn[['D', 'T', 'LSreturn']]
# df_lsreturn.columns = ['D', 'T', col_name]
#
#
# # df_lsreturn.to_csv("lsreturn.csv")
# """4.7 Longshort部分之Summary模块"""
#
# with pd.ExcelWriter('output.xlsx') as writer:
#     df_ic.to_excel(writer, sheet_name='ICs')
#     df_coverage.to_excel(writer, sheet_name='coverage')
#     df_turnover.to_excel(writer, sheet_name='turnover')
#     df_lsreturn.to_excel(writer, sheet_name='LSreturns')
#     df_hitrate.to_excel(writer, sheet_name='HitRate')
#     df_wealth.to_excel(writer, sheet_name='wealth')
#

##############################
# 计算summary部分
##############################

"""
def max_drawdown(xs: list) -> float:
    mdd = 0
    maxwealth = 1
    x = iter(xs)
    while True:
        try:
            value = next(x)
        except Exception as e:
            return mdd
        maxwealth = max(maxwealth, value)
        mdd = min(mdd, value/maxwealth - 1)


df = pd.read_csv("./output/DTech_amt_single.csv")
df = df.sort_values(by=['D', 'T'], axis=0, ascending=True)

N = df.shape[0]

FIRST_DAY = 0
LAST_ONE_DAY = -1
LAST_ONE_MONTH = -22
LAST_THREE_MONTH = -64
LAST_ONE_YEAR = -245
LAST_THREE_YEAR = -490
LAST_FIVE_YEAR = -1225
DAYS_PER_YEAR = 244

summary = {}
"""

"""
低于一年是的
1~3年的话  cagr=cagr3 = cagr5
3~5年的话 cagr=cagr5
"""

"""
btm_w = df.iloc[0, df.columns.get_loc('wealth')]
btm_d1 = df.iloc[LAST_ONE_DAY, df.columns.get_loc('wealth')]
btm_m1 = df.iloc[LAST_ONE_MONTH, df.columns.get_loc('wealth')]
btm_m3 = df.iloc[LAST_THREE_MONTH, df.columns.get_loc('wealth')]
btm_y1 = df.iloc[LAST_ONE_YEAR, df.columns.get_loc('wealth')]
btm_y3 = df.iloc[LAST_THREE_YEAR, df.columns.get_loc('wealth')]
btm_y5 = df.iloc[LAST_FIVE_YEAR, df.columns.get_loc('wealth')]

return_Z = btm_d1/btm_w - 1
summary['return_1m'] = (btm_d1 - 1) if N < abs(LAST_ONE_MONTH) else (btm_d1/btm_m1 - 1)
summary['return_3m'] = (btm_d1 - 1) if N < abs(LAST_THREE_MONTH) else (btm_d1/btm_m3 - 1)
return_Z1y = (btm_d1 - 1) if N < abs(LAST_ONE_YEAR) else (btm_d1/btm_y1 - 1)
return_Z3y = (btm_d1 - 1) if N < abs(LAST_THREE_YEAR) else (btm_d1/btm_y3 - 1)
return_Z5y = (btm_d1 - 1) if N < abs(LAST_FIVE_YEAR) else (btm_d1/btm_y5 - 1)

summary['cagr'] = (return_Z + 1) ** (DAYS_PER_YEAR / N) - 1
summary['cagr_1y'] = summary['cagr'] if N < abs(LAST_ONE_YEAR) else return_Z1y
summary['cagr_3y'] = summary['cagr'] if N < abs(LAST_THREE_YEAR) else ((return_Z3y + 1) ** (DAYS_PER_YEAR / abs(LAST_THREE_YEAR)) - 1.0)
summary['cagr_5y'] = summary['cagr'] if N < abs(LAST_FIVE_YEAR) else ((return_Z5y + 1) ** (DAYS_PER_YEAR / abs(LAST_FIVE_YEAR)) - 1.0)
summary['volatility'] = df['LSreturn'].std(ddof=1) * np.sqrt(DAYS_PER_YEAR)
summary['sharpe_ratio'] = summary['cagr'] / summary['volatility']
summary['rank_ic_mean'] = df['ic'].mean()
summary['rank_ic_std'] = df['ic'].std(ddof=1)
summary['rank_ic_std'] = summary['rank_ic_mean'] / summary['rank_ic_std']
summary['hit_rate'] = df['hitrate'].mean() * 100.0
summary['max_drawdown'] = max_drawdown(df['wealth'].tolist())

summary['turnover'] = df['turnover'].mean() * 100
"""



"""
CAGR算法

对于X1年：获取X1年 最后一个交易日 4.6中计算得到的Wealth值W 返回W ** （M/252）- 1, M为当年加入计算的交易日天数

对于XN年，分别获取回测期内 XN-1年 最后一个交易日和XN年最后一个交易日（XN年最后一个交易日不一定是当年真正的最后一个交易日，因为回测期没有到XN年的末尾） 4.6中计算得到的Wealth值W1，W2，返回（W2/W1） ** （M/252）- 1, M为当年加入计算的交易日天数

对于其他年份y，分别获取 y-1年 最后一个交易日和y年最后一个交易日4.6中计算得到的Wealth值W1，W2，返回（W2/W1）- 1
"""

"""
df['Y'] = df['D'].apply(lambda day: str(day)[0:4])
df_cagr_agg = df.groupby(by=['Y'], as_index=False).agg(M=('D', 'count'), D_min=('D', 'min'), D_max=('D', 'max'))
df_cagr_agg = pd.merge(df_cagr_agg, df[['D', 'wealth']], left_on='D_min', right_on='D')
df_cagr_agg = df_cagr_agg.rename(columns={"wealth": "D_min_wealth"})
df_cagr_agg.drop('D', axis=1, inplace=True)
df_cagr_agg = pd.merge(df_cagr_agg, df[['D', 'wealth']], left_on='D_max', right_on='D')
df_cagr_agg = df_cagr_agg.rename(columns={"wealth": "D_max_wealth"})
df_cagr_agg.drop('D', axis=1, inplace=True)
df_cagr_agg['D_max_wealth_1'] = df_cagr_agg['D_max_wealth'].shift(1)



df['Y'] = df['D'].apply(lambda day: str(day)[0:4])
df_cagr_agg = df.groupby(by=['Y'], as_index=False).agg(M=('D', 'count'), D_max=('D', 'max'))
df_cagr_agg = pd.merge(df_cagr_agg, df[['D', 'wealth']], left_on='D_max', right_on='D')
df_cagr_agg = df_cagr_agg.rename(columns={"wealth": "D_max_wealth"})
df_cagr_agg.drop('D', axis=1, inplace=True)
df_cagr_agg['D_max_wealth_1'] = df_cagr_agg['D_max_wealth'].shift(1)
for _, row in df_cagr_agg.iterrows():
    cagr_year_key = "cagr_year_{}".format(row['Y'])
    if str(row['D_max_wealth_1']) == 'nan':
        summary[cagr_year_key] = row['D_max_wealth'] ** (row['M']/float(DAYS_PER_YEAR)) - 1.0
    else:
        summary[cagr_year_key] = (row['D_max_wealth']/row['D_max_wealth_1']) ** (row['M']/float(DAYS_PER_YEAR)) - 1.0

df_ic_agg = df.groupby(by=['Y'], as_index=False).agg(rank_ic_mean_year=('ic', 'mean'))
for _, row in df_ic_agg.iterrows():
    summary["rank_ic_mean_year_{}".format(row['Y'])] = row['rank_ic_mean_year']

"""
from fatbulls.context import FatbullsFactorContext

class FactorCalculateTask(object):
    """多线程执行引擎"""

    def __init__(self, factor_name: str, factor_data: dict, forward_return: pd.DataFrame):
        self.factor_name = factor_name
        self.factor_data = factor_data
        self.forward_return = forward_return
        self._factor_config = FatbullsFactorContext.get_instance().get_factor_config()
        self.forward_return_name = list(forward_return.columns)[-1]
        self.factor_forward = self.factor_name + '_' + self.forward_return_name
        self.grouped = None
        self.df_q = None
        self.df_backtest_res = None
        self.backtest_result = dict()

    def __call__(self, **kwargs):
        self._middle_calculate()
        self._calculate_ic()
        self._calculate_coverage()
        self._calculate_turnover()
        self._calculate_lsreturn_and_wealth()
        self._calculate_hitrate()
        self.df_backtest_res.to_csv("./output/{}_single.csv".format(self.factor_name))
        self._clean()
        return self.df_backtest_res

    def _calc_turnover(self, x, y):
        """计算交集"""
        if x is None or y is None:
            return 1.0
        return 1.0 - len(set(x).intersection(set(y))) / (len(x) + len(y))

    def _calc_hitrate(self, group_df):
        """计算HitRate"""
        row_size = group_df.shape[0]
        if row_size <= 0:
            return 0.0
        group_df_q1 = group_df[group_df.q == 'Q1']
        group_df_q10 = group_df[group_df.q == 'Q10']
        arr_pos = np.array(group_df_q1[self.forward_return_name])
        arr_neg = np.array(group_df_q10[self.forward_return_name])
        pos_count = np.sum(arr_pos > 0)
        neg_count = np.sum(arr_neg < 0)
        return float(pos_count + neg_count) / row_size

    def _middle_calculate(self):
        """中间计算：单因子加载所有csv文件"""
        df_factor_data = pd.concat(self.factor_data.values(), axis=0, ignore_index=True)
        df_all = pd.merge(df_factor_data, self.forward_return, on=['K', 'D', 'T'], how='left')
        df_all = df_all.dropna(subset=[self.factor_name])
        df_all = df_all.dropna(subset=[self.forward_return_name])
        self.grouped = df_all.groupby(by=['D', 'T'], as_index=False)
        df_all['q'] = df_all.groupby(by=['D', 'T'], as_index=False)[self.factor_name].transform(
            lambda x: pd.qcut(x, 10, labels=["Q{}".format(i) for i in range(1, 11)]))
        df_all['q'] = df_all['q'].astype(str)
        self.df_coverage_summary = df_all.groupby(by=['D', 'T'], as_index=False)
        self.df_q = df_all.query("q == 'Q1' or q == 'Q10'")

    def _clean(self):
        del self.df_q
        del self.grouped

    def _calculate_ic(self):
        """4.1 计算IC"""
        df_ic = self.grouped.apply(lambda x: x[self.factor_name].corr(x[self.forward_return_name], 'spearman'))
        df_ic.columns = ['D', 'T', 'ic']
        self.backtest_result['ic'] = df_ic
        self.df_backtest_res = df_ic

    def _calculate_coverage(self):
        """4.2 LongShort部分之计算覆盖率"""
        df_coverage = self.grouped.agg({'K': 'count'})
        df_coverage = df_coverage.rename(columns={"K": 'coverage'})
        df_coverage = df_coverage[['D', 'T', 'coverage']]
        self.backtest_result['coverage'] = df_coverage
        # self.df_backtest_res = pd.merge(self.df_backtest_res, df_coverage, on=['D', 'T'], how='left')

    def _calculate_turnover(self):
        """4.3 LongShort部分之计算换手率"""
        df_turnover = self.df_q.groupby(['D', 'T'], as_index=False).agg({"K": list})
        df_turnover.columns = ['D', 'T', 'K_list']
        df_turnover['K_list_prev'] = df_turnover.sort_values(by=['D', 'T'], ascending=True)['K_list'].shift(1)
        df_turnover['turnover'] = df_turnover.apply(lambda x: self._calc_turnover(x['K_list'], x['K_list_prev']),
                                                    axis=1)
        df_turnover = df_turnover[['D', 'T', 'turnover']]
        self.backtest_result['turnover'] = df_turnover
        # self.df_backtest_res = pd.merge(self.df_backtest_res, df_turnover, on=['D', 'T'], how='left')

    def _calculate_lsreturn_and_wealth(self):
        """4.4 LongShort部分之计算LSreturn And Wealth"""
        df_lsreturn = self.df_q.groupby(by=['D', 'T', 'q'], as_index=False)[self.forward_return_name].mean()
        df_lsreturn = df_lsreturn.pivot(index=['D', 'T'], columns='q', values=self.forward_return_name).reset_index()
        df_lsreturn['LSreturn'] = df_lsreturn.apply(lambda row: row['Q1'] - row['Q10'], axis=1)
        df_lsreturn['wealth'] = df_lsreturn['LSreturn'].apply(lambda x: x + 1).cumprod()
        df_lsreturn = df_lsreturn[['D', 'T', 'LSreturn', 'wealth']]
        self.backtest_result['lsreturn'] = df_lsreturn
        # self.df_backtest_res = pd.merge(self.df_backtest_res, df_lsreturn, on=['D', 'T'], how='left')

    def _calculate_hitrate(self):
        """4.5 LongShort部分之计算HitRate"""
        df_hitrate = self.df_q.groupby(by=['D', 'T'], as_index=False).apply(self._calc_hitrate)
        df_hitrate.columns = ['D', 'T', 'hitrate']
        self.backtest_result['hitrate'] = df_hitrate
        # self.df_backtest_res = pd.merge(self.df_backtest_res, df_hitrate, on=['D', 'T'], how='left')

    def _get_summary(self):
        pass

    def _graph_ic_month(self):
        pass

    def _graph_ic_industry(self):
        pass

    def _graph_hit_rate_10bin(self):
        pass

    def _sf_report(self):
        pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: run_concurrent.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/18 12:45:26
"""
import time
import random

from concurrent.futures import ThreadPoolExecutor
from typing import Generator, Any, Optional, Callable, Tuple


def task_1(path):
    time.sleep(random.randint(1, 5))
    print("task_1")
    print(path)


def task_2(path):
    time.sleep(random.randint(1, 5))
    print("task_2")
    print(path)


def task_3(path):
    time.sleep(random.randint(1, 5))
    print("task_3")
    print(path)


def task_4(path):
    time.sleep(random.randint(1, 5))
    print("task_4")
    print(path)


gen_funcs = (
    task_1, task_2, task_3, task_4
)


class FactorCalculateTask(object):

    def __init__(self, factor):
        self._factor = factor

    def __call__(self, path, fields, **kwargs):
        print("process factor: {}".format(self._factor))
        ret = random.randint(1, 5)
        time.sleep(ret)
        return ret

    def _calculate_ic(self):
        """计算IC"""
        pass

    def _calculate_hitrate(self):
        """计算胜率"""
        pass

    def _calculate_return(self):
        """计算收益、仅3个月、6个月、1年、3年、5年收益率和CAGR"""
        pass

    def _calculate_wealth(self):
        """计算净值"""
        pass

    def _calculate_turnover(self):
        """计算换手率"""
        pass

    def _calculate_coverage(self):
        """计算覆盖率"""
        pass

    def _get_summary(self):
        pass

    def _graph_ic_month(self):
        pass

    def _graph_ic_industry(self):
        pass

    def _graph_hit_rate_10bin(self):
        pass

    def _sf_report(self):
        pass


other = dict()
other['compression'] = 9
with ThreadPoolExecutor(max_workers=2) as executor:
    futures = []
    for fac in ['factor_{}'.format(i) for i in range(1, 20)]:
        future = executor.submit(FactorCalculateTask(fac), "/data/quant", "['', '']", **other)
        futures.append(future)
    executor.shutdown(wait=True)
    for future in futures:
        print(future.result())






#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: run_factor_backtest.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/15 09:50:03
"""
import sys
import os
sys.path.append('..')

import fatbulls
from pathlib import Path


# output_path = os.path.join(Path(__file__).absolute().parent.parent, "output")
output_path = "/data/quant/report/factor_backtest/"

config = {
    'factor': {
        'base_report_dirpath': output_path,   # 可选, 因子报告输出路径
        'backtest_name': 'test_backtest',  # 可选，回测名称，可以不唯一
        'sd': '20160101',    # 回测区间，开始时间
        'ed': '20230606',  # 回测区间，结束时间
        'task_num': '101',  # 回测任务ID, 唯一，调用方传入
        'univ_name': 'cnall',      # 可选，资产集名称，默认cnall, 包括所有股票
        'fdays': 1,                # 可选     # 未来N天的收益来做Y
        'rettype': 'vwap30',       # 可选, 收益率的类型，比如：vwap30为开盘前30分钟的成交量加权的平均价
        'factor_dirs': [
            # '/data/quant/factors/DTech/DTech_volume',  # 因子路径，最后一个目录代表因子名，内部为 {DAY}.csv
            '/data/quant/factors/DTech/DTech_amt',
            '/data/quant/factors/DTech/DTech_high',
            '/data/quant/factors/DTech/DTech_ret',
            '/data/quant/factors/DTech/DTech_opn',
            '/data/quant/factors/DTech/DTech_vwap'
        ],
        "base_log_path": "/data/user/wangjiangfeng/fatbulls"
    }
}

fatbulls.run_factor_backtest(config=config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# Copyright (c) 2023 datavita.com.cn, Inc. All Rights Reserved
#
########################################################################


"""
File: main.py
Author: wangjiangfeng(wangjiangfeng@hcyjs.com)
Date: 2023/8/14 17:30:51
"""

import logging
import pandas as pd
from .context import FatbullsFactorContext
from .utils import init_fqdatac_env
from .utils.logger import init_logging
from .factor.back_test import FactorBackTestEngine
from .factor.back_test import BackTestConfig
from .factor.se import fac_backtest


def set_pandas_options():
    pd.set_option("display.max_columns", 100)
    pd.set_option("display.width", 1000)
    pd.set_option("display.max_colwidth", 100)
    pd.set_option("display.float_format", lambda x: "%.4f" % x)


def init_fqdatac(fqdatac_uri):
    if fqdatac_uri in ["disabled", "DISABLED"]:
        return
    try:
        import fqdatac
    except ImportError:
        return
    if isinstance(fqdatac.client.get_client(), fqdatac.client.DummyClient):
        init_fqdatac_env(fqdatac_uri)
        try:
            fqdatac.init()
        except Exception as e:
            logging.warning('fqdatac init failed, some apis will not function properly: {}'.format(str(e)))


def _execute_factor_backtest(backtest_config: BackTestConfig, report_path: str = ''):
    fac_backtest.run(backtest_config.backtest_name,
                     sd=backtest_config.sd,
                     ed=backtest_config.ed,
                     task_num=backtest_config.task_num,
                     univ_name=backtest_config.univ_name,
                     fdays=backtest_config.fdays,
                     rettype=backtest_config.rettype,
                     directory_list=backtest_config.factor_dirs,
                     output_path=report_path
    )


def run(config=None, app_name='fatbulls_factor_backtest', user_funcs=None, set_pandas=True):
    try:
        if set_pandas:
            """pandas 设置"""
            set_pandas_options()
        if config.factor.base_log_path:
            app_dir = config.factor.base_log_path
        else:
            app_dir = None
        init_logging(app_name, app_dir)
        init_fqdatac(getattr(config.base, 'fqdatac_uri', None))
        fat_ctx = FatbullsFactorContext(config)
        factor_conf = fat_ctx.get_factor_config()
        fat_ctx.get_base_report_dirpath()
        logging.info("---start factor evaluation---")
        _execute_factor_backtest(factor_conf, fat_ctx.get_base_report_dirpath())
        logging.warning(f"All done")
        # engine = FactorBackTestEngine(factor_conf)
        # """第一步初始化"""
        # engine.initialize()
        # """数据加载与预处理"""
        # engine.load_factors()
        # """执行回测逻辑"""
        # engine.run(max_workers=10)
        #
        #
        # # 整个回测过程
        # logging.info("start factor evaluation")
        # evaluator = FactorEvaluator(
        #     sd=loader.sd,
        #     ed=loader.ed,
        #     backtest_name=loader.backtest_name,
        #     univ_name=loader.univ_name,
        #     fdays=loader.fdays,
        #     rettype=loader.rettype,
        #     directory_list=loader.directory_list,
        #     raw=loader.container,
        #     fwdret=loader.fwdret1,
        #     output_path=factor_conf.base_report_dirpath,
        #     task_num=factor_conf.task_num
        # )
        # evaluator.initialize()
        # evaluator.nextfactor()
        # while evaluator.whole_processing:
        #     logging.info(f"factor evaluation for factor{evaluator.curfac}")
        #     evaluator.evaluate()
        #     evaluator.fac_report()
        #     evaluator.nextfactor()
        # evaluator.output()
    except Exception as e:
        raise e


